/* -*-mode: C;  tab-width: 2; -*- */
%{
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <mysql.h>
#include <malloc.h>

int devel=0;
FILE *indexfile=0;
FILE *titlefile=0;
int index_requested=0;
int verbose=0;
MYSQL *connection=NULL;
char *mysql_host=NULL;
char *mysql_database=NULL;
char *mysql_username=NULL;
char *mysql_password=NULL;
char *include_dir=NULL;

struct user_var
{
  struct user_var* next;
  char* name;
  char* value;
};

struct user_var* user_vars=NULL;

#define HELPMSG "Usage: yahp [-i index file] [-t title file] [-I include dir]\
[-h mysql_host] [-m mysql_database] [-u mysql_user] [-p mysql_password] [-d] <template file>\n"

#define MAX_INCLUDE_DEPTH 10
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_state[MAX_INCLUDE_DEPTH];
int include_stack_ptr=0;


int my_strnlen(const char * inp_str,int max_len)
{
  int loop;
  for(loop=0;loop<=max_len;loop++)
		if(inp_str[loop]==0)
			break;
	return loop;
}

char * my_strndup(const char * inp_str,int max_len)
{
	char * new_str;
	int inp_str_len;
	inp_str_len=my_strnlen(inp_str,max_len);
	if(inp_str_len<max_len)
		max_len=inp_str_len;
	new_str=malloc(max_len+1);
	strncpy(new_str,inp_str,max_len);
	new_str[max_len]=0;
	return new_str;
}

struct user_var* find_uservar(const char * findname)
{
  struct user_var* var;
	var=user_vars;
	while(var!= NULL)
	{
		if(strcasecmp(var->name,findname)==0)
			return var;
		var=var->next;
	}
	return NULL; /* failed to find var */
}

struct user_var* set_uservar(char * name,char *value)
{
	struct user_var* var;
	var=find_uservar(name);
	if(var==NULL)
	{
		/* make new var */
		var=malloc(sizeof(struct user_var));
		var->name=strdup(name);
		var->value=NULL;
		var->next=user_vars;
		user_vars=var;
	}

	if(var->value!=NULL)
		free(var->value);

	var->value=strdup(value);

	return var;
}

int do_uservar(const char * assign)
{
  char * equals;
	char * name;
  struct user_var* var;
  equals=index(assign,'=');

  if(equals!=NULL)
	{
		/* assign variable (possibly creates new one) */
		name=my_strndup(assign,(equals-assign));
		var=set_uservar(name,&equals[1]);
		free(name);
	}
  else
	{
		var=find_uservar(assign);
		if(var!=NULL)
			fprintf(yyout, var->value);
		else
			fprintf(stderr,"Reference to undefined variable (%s)\n",assign);
  }
	return 0;
}
%}

%x comment
%x fileinc
%x ignore
%x title
%x uservar
%x setp
%x mysql
%%

"<!--" BEGIN(comment);
<comment>"_ph_date" { time_t T=time(NULL);
											char *t = ctime(&T);
											char *s = (char*)malloc(16+strlen(t));
											strcpy(s, "Last modified: "); strcat(s, t);
											fprintf(yyout, s);
											if(index_requested && indexfile) fprintf(indexfile, s);
											free(s); // free(t);
										}
<comment>"_ph_title" 	{ struct user_var* var;
												var=find_uservar("pagetitle");
												if(var){
													fprintf(yyout, var->value);
													if(index_requested && indexfile) fprintf(indexfile, var->value);
												}
											}
<comment>"_ph_inc"([ \t\n]*)\" BEGIN(fileinc);
<comment>"_ph_dev"([ \t\n]*)\" { if(devel) BEGIN(fileinc); else BEGIN(ignore); }
<comment>"_ph_index" { index_requested=(index_requested+1)%2; }
<comment>"_ph_pagetitle"([ \t\n]*)\" BEGIN(title);
<comment>"_ph_var"([ \t\n]*)\" BEGIN(uservar);
<comment>"_ph_mysql"([ \t\n]*)\" BEGIN(mysql);
<comment>"-->" BEGIN(INITIAL);
<comment>(.|\n) /* ignore */;

<mysql>[^"\n]*	{ char query[62];
									if(!connection){
										connection = mysql_init(NULL);
										if(!mysql_real_connect(connection, mysql_host, mysql_username, mysql_password, mysql_database, 0, NULL, 0)){
											fprintf(stderr, "Error connecting to database %s:%s:%s\n", mysql_host, mysql_database, mysql_username);
											exit(1);
										}
									}
									// We now have a connection
									// Code here for getting what we want
									strcpy(query, "SELECT content FROM htmlblocks WHERE name='");
									strncat(query, yytext, 16);
									strcat(query, "'");
									if(mysql_query(connection, query)==0){ // If successfull
										MYSQL_RES *res;
										res = mysql_store_result(connection);
										if(mysql_num_rows(res)>0){
											MYSQL_ROW row;
											char *s; int i;
											if(verbose) fprintf(stderr, "Successfully retrieved text for %s.\n", yytext);
											row = mysql_fetch_row(res);
											s=row[0];
											for(i=0; i<strlen(s); i++){
												if(s[i]!=13) putchar(s[i]); // Don't output ^M's that might have snuck in
											}
										} else { // No rows returned!
											fprintf(stderr, "No mysql row called %s\n", yytext);
											exit(1);
										}
									}
								}
<mysql>\"	BEGIN(comment);

<title>[^"\n]* {  struct user_var* var;
                  var=set_uservar("pagetitle",yytext);
								  if(verbose)
										fprintf(stderr, "Found page title: %s\n", var->value);
								  if(titlefile)
										fprintf(titlefile, var->value);
								}
<title>\" BEGIN(comment);

<uservar>[^"\n]* {
  do_uservar(yytext);
								}
<uservar>\" BEGIN(comment);

<fileinc>[^"\n]* { /* include file */
	char * filename;

  if(include_stack_ptr>=MAX_INCLUDE_DEPTH)
  {
    fprintf(stderr,"Includes nested too deeply");
    exit(1);
  }

  /* save both current buffer and start state for later restore */
  include_stack[include_stack_ptr]=YY_CURRENT_BUFFER;
  include_state[include_stack_ptr]=YY_START;
  include_stack_ptr++;

	if(verbose) fprintf(stderr, "Include file: %s\n", yytext);
	yyin = fopen(yytext, "r");
	if((!yyin) && (include_dir!=NULL)){ /* try include dir */
  		filename=malloc(1024);
  		snprintf(filename,1024,"%s/%s",include_dir,yytext);
			yyin = fopen(filename, "r");
  		free(filename);
	}
	if(!yyin){
		fprintf(stderr, "Bad include file: %s", yytext);
		exit(1);
	}

  yy_switch_to_buffer(yy_create_buffer(yyin,YY_BUF_SIZE));

  BEGIN(INITIAL);

}
<fileinc>\"	BEGIN(comment);

<ignore>[^"\n]* { if(verbose) fprintf(stderr, "Ignore include file: %s\n", yytext); }
<ignore>\"	BEGIN(comment);
.|\n { ECHO; if(index_requested && indexfile){ fprintf(indexfile, "%s", yytext);} }

<<EOF>> {
    if(--include_stack_ptr<0)
    {
	yyterminate();
    }
    else
    {
        yy_delete_buffer(YY_CURRENT_BUFFER);
        yy_switch_to_buffer(include_stack[include_stack_ptr]);
	BEGIN(include_state[include_stack_ptr]);
    }
}

%%

int main(int argc, char **argv){
	int result;
	while((result=getopt(argc, argv, "+di:t:h:m:u:p:I:v"))!=EOF){
		switch(result){
			case 'd': devel=1; break;
			case 'i': indexfile=fopen(optarg, "w"); break;
			case 't': titlefile=fopen(optarg, "w"); break;
			case 'h': mysql_host=strdup(optarg); break;
			case 'm': mysql_database=strdup(optarg); break;
			case 'u': mysql_username=strdup(optarg); break;
			case 'p': mysql_password=strdup(optarg); break;
			case 'I': include_dir=strdup(optarg); break;
			case 'v': verbose=1; break;
			default: { printf(HELPMSG); exit(1); }
		}
	}
	if(argc > optind){
		yyin=fopen(argv[optind], "r");
	} else {
		yyin=stdin;
	}
	yylex();
	if(indexfile){
		fclose(indexfile);
	}
	if(titlefile){
		fclose(titlefile);
	}
	if(connection){
		mysql_close(connection);
	}
	return 0;
}

// vim: ts=2
